#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script handling 'BreathingMovement' object: a discretized chest/abdominal movement.

Created on Fri Sep  5 08:12:19 2025
@author: CoustilletT
"""


import pandas as pd
from scipy.signal import detrend

from . import featureextraction as features
from .instantiationmethods import _from_file, _from_dataframe, _process_time
from . import visualization
from .utils import enforce_type_arg, create_absolute_time, ComparableMixin


class BreathingMovement(ComparableMixin):
    """Breathing movement: thorax or abdomen."""

    def __init__(self, time, movements, movement_type, identifier, detrend_y=False):
        self.time = time
        self.movements = movements
        self.movement_type = movement_type
        self.identifier = identifier

        self.processed_time = _process_time(time)

        self.absolute_time = create_absolute_time(self.time, self.get_hz())

        if detrend_y:
            self.movement = detrend(self.raw_movements, type="constant")

    from_file = classmethod(_from_file)
    from_dataframe = classmethod(_from_dataframe)

    def __getitem__(self, key):
        """
        To allow a 'BreathingMovement' object to be sliced and used as a new object.

        Args:
        ----
            key (list): slice of shape [start:stop:steps].

        Returns:
        -------
            A new 'BreathingMovement' sliced object.

        """
        sliced_object = self.__class__(
            time=self.time[key],
            movements=self.movements[key],
            movement_type=self.movement_type,
            identifier=self.identifier,
            detrend_y=False
        )

        return sliced_object

    def get_hz(self):
        """To get the sampling rate of the discretized breathing signal."""
        return features.compute_sampling_rate(x=self.processed_time)

    def get_frequency(self):
        """
        Get movement frequency.
        """
        freq = features.frequency(
            signal=self.movements,
            sampling_rate=self.get_hz(),
            method="welch",
            which_peaks=None,
            distance=-1,
            n_digits=3,
        )

        return freq

    @enforce_type_arg(shape=str)
    def get_info(self, shape="dict"):
        """To get signal information (identifier, start, end and duration.)"""
        info = {
            "identifier": self.identifier,
            "start": self.time[0],
            "end": self.time[-1],
            "duration": str(pd.to_timedelta(
                self.absolute_time[-1], unit="s")
            ).split()[-1]
        }

        if shape == "dict":
            return info
        if shape == "df":
            return pd.DataFrame([info])
        raise ValueError(f"shape must be 'dict' or 'df', not '{shape}'.")

    @enforce_type_arg(output_path=str)
    def plot(self, output_path=""):
        visualization.plot_signal(
            x=self.absolute_time,
            y=self.movements,
            show_zeros=False,
            show_segments=False,
            show_auc=False,
            highlight_time=(),
            highlight_auc=(),
            label=f"{self.movement_type} movements",
            output_path=output_path,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script handling the instantiation methods for the objects.

Created on Mon Sep  8 10:13:21 2025
@author: CoustilletT
"""


from importlib.resources import files, as_file
import pandas as pd
import re
from .utils import enforce_type_arg, print_source


TIME_PATTERN_1 = r"^\d{2}:\d{2}:\d{2}([.:])\d+$"
TIME_PATTERN_2 = r"^[+-]?\d+([.,]\d+)?([eE+][+-]?\d+)?$"
TIME_PATTERN = f"(?:{TIME_PATTERN_1})|(?:{TIME_PATTERN_2})"


def process_dataframe(raw_data):
    """
    To process the raw dataframe and properly format it for object instantiation.

    Args:
    ----
        raw_data (pd.DataFrame): Raw dataframe = output of plethysmography software.

    Returns:
    -------
        data (pd.DataFrame): Processed DataFrame = correctly formatted DataFrame.

    """
    raw_data = raw_data.astype(str)
    data = raw_data[raw_data["time"].str.match(TIME_PATTERN, na=False)].reset_index(
        drop=True
    )
    if not data["time"].str.contains(":").any():
        data["time"] = data["time"].str.replace(",", ".")
    if data["values"].str.contains(",").any():
        data["values"] = data["values"].str.replace(",", ".")

    data["values"] = data["values"].astype(float)

    return data


def _process_time(time_vector):
    """
    To process the time_vector and properly format it for object instantiation.

    Args:
    ----
        time_vector (array): time vector associated with discretized air flow.

    Returns:
    -------
        array: Processed time vector = correctly formatted time_vector.

    """
    time_vector = pd.Series(time_vector, dtype=str)

    is_wrong_format = time_vector.str.match(r"^\d{2}:\d{2}:\d{2}:\d+$").all()
    if is_wrong_format:
        time_vector = time_vector.str.replace(r":(?=\d+$)", ".", regex=True)

    # To instantiate an object even if the time vector is not in absolute seconds.
    # Required format : HH:MM:SS.XXX
    if not all(
        re.match(TIME_PATTERN_2, time) for time in time_vector.values.astype(str)
    ):
        time_vector = pd.to_timedelta(time_vector).dt.total_seconds()

    time_vector = time_vector.astype(float)

    return time_vector.values


@enforce_type_arg(identifier=str, filename=str, sheet_name=str, detrend_y=bool, movement_type=str)
def _from_file(cls, identifier, filename, sheet_name="", detrend_y=False, movement_type=""):
    """
    To instantiate a 'BreathingFlow' or 'BreathingMovement' objet from a file path.

    Args:
    ----
        identifier (str): breathing signal identifier.
        filename (str): path to the two-column file representing
                        discretized time and discretized air flow rate/movements.
        sheet_name (str, optional): name of the sheet to be considered.
                                    Defaults to "".
        detrend_y (bool, optional): to set the mean of the air flow rate at 0.
                                    Defaults to False.
        movement_type (str, optional): type of breathing movements (thorax/abdomen),
                                       if any. Defaults to "".

    Returns:
    -------
        BreathingFlow: instantiate an objet of type 'BreathingFlow'.
        OR
        BreathingMovement: instantiate an objet of type 'BreathingMovement'.

    """
    col_names = ["time", "values"]
    match filename:
        case _ if filename.endswith("txt"):
            raw_data = pd.read_csv(
                filename, sep=r"\s+", usecols=[0, 1], names=col_names, dtype=str
            )
        case _ if filename.endswith(("xlsx", "xls")):
            sheet_name = sheet_name if sheet_name else 0
            raw_data = pd.read_excel(
                filename, names=col_names, sheet_name=sheet_name, dtype=str,
                header=None
            )

    data = process_dataframe(raw_data)

    if cls.__name__ == "BreathingFlow":
        if movement_type:
            raise ValueError(
                "A 'BreathingFlow' object cannot have a 'movement_type' argument. "
                "Use a 'BreathingMovement' object to specify the type of movements."
            )
        return cls(
                identifier=identifier,
                raw_time=data["time"].values,
                raw_flow=data["values"].values,
                detrend_y=detrend_y,
            )

    if cls.__name__ == "BreathingMovement":
        if not movement_type:
            raise TypeError("missing a required argument: 'movement_type'")
        return cls(
            identifier=identifier,
            time=data["time"].values,
            movements=data["values"].values,
            movement_type=movement_type,
            detrend_y=detrend_y,
        )


@enforce_type_arg(identifier=str, detrend_y=bool, movement_type=str)
def _from_dataframe(cls, identifier, df, detrend_y=False, movement_type=""):
    """
    To instantiate a 'BreathingFlow' objet from a dataframe.

    Args:
    ----
        identifier (str): breathing signal identifier.
        df (pandas.DataFrame): two-column dataframe representing discretized
                               time and discretized air flow rate.
        detrend_y (bool, optional): to set the mean of the air flow rate at 0.
                                    Defaults to False.
        movement_type (str, optional): type of breathing movements (thorax/abdomen),
                                       if any. Defaults to "".

    Returns:
    -------
        BreathingFlow: instantiate an objet of type 'BreathingFlow'.
        OR
        BreathingMovement: instantiate an objet of type 'BreathingMovement'.

    """
    if not {"time", "values"}.issubset(df.columns):
        raise ValueError(
            "DataFrame must contain a 'time' column and a 'values' column."
        )

    data = process_dataframe(df)

    if cls.__name__ == "BreathingFlow":
        if movement_type:
            raise ValueError(
                "A 'BreathingFlow' object cannot have a 'movement_type' argument. "
                "Use a 'BreathingMovement' object to specify the type of movements."
            )
        return cls(
            identifier=identifier,
            raw_time=data["time"].values,
            raw_flow=data["values"].values,
            detrend_y=detrend_y,
        )

    if cls.__name__ == "BreathingMovement":
        if not movement_type:
            raise TypeError("missing a required argument: 'movement_type'")
        return cls(
            identifier=identifier,
            time=data["time"].values,
            movements=data["values"].values,
            movement_type=movement_type,
            detrend_y=detrend_y,
        )


def _load_sinus(cls):
    """
    Load and return a BreathingFlow object from the "sinus" dataset.

    Returns:
    -------
        BreathingFlow: a BreathingFlow object = the sinus function.

    Note:
    ----
        is used to demonstrate the 'proof of concept'.
    """
    sinus_resource = files("pybreathe.datasets").joinpath("sinus.txt")

    with as_file(sinus_resource) as sinus_path:
        with open(sinus_path, encoding="utf-8") as f:
            sinus = pd.read_csv(
                f, sep="\t", names=["time", "values"], dtype=float
            )

    return cls(
        identifier="example_sinus",
        raw_time=sinus["time"].values,
        raw_flow=sinus["values"].values,
        detrend_y=False,
    )


def _load_breathing_like_signal_01(cls):
    """
    Load and return a BreathingFlow object from the "breathing-like signal 01" dataset.

    Returns:
    -------
        BreathingFlow: a BreathingFlow object.
    """
    breathing_resource_01 = (
        files("pybreathe.datasets")
        .joinpath("breathing_like_signal_01.txt")
    )

    with as_file(breathing_resource_01) as breathing_01_path:
        with open(breathing_01_path, encoding="utf-8") as f:
            breathing_01 = pd.read_csv(
                f, sep="\t", names=["time", "values"], dtype=float
            )

    print(print_source()["breathing-like 01"])

    return cls(
        identifier="example_breathing-like_signal_01",
        raw_time=breathing_01["time"].values,
        raw_flow=breathing_01["values"].values,
        detrend_y=False,
    )


def _load_breathing_like_signal_02(cls):
    """
    Load and return a BreathingFlow object from the "breathing-like signal 02" dataset.

    Returns:
    -------
        BreathingFlow: a BreathingFlow object.
    """
    breathing_resource_02 = (
        files("pybreathe.datasets")
        .joinpath("breathing_like_signal_02.txt")
    )

    with as_file(breathing_resource_02) as breathing_02_path:
        with open(breathing_02_path, encoding="utf-8") as f:
            breathing_02 = pd.read_csv(
                f, sep="\t", names=["time", "values"], dtype=float
            )

    print(print_source()["breathing-like 02"])

    return cls(
        identifier="example_breathing-like_signal_02",
        raw_time=breathing_02["time"].values,
        raw_flow=breathing_02["values"].values,
        detrend_y=False,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions that handle plots.

Created on Wed Apr  9 08:30:59 2025
@author: CoustilletT
"""


import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import ScalarMappable
from matplotlib.collections import LineCollection
import matplotlib.colors as mcolors
from matplotlib.lines import Line2D
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from scipy.integrate import trapezoid
from scipy.signal import find_peaks
import seaborn as sns

from . import featureextraction as features
from .utils import scientific_round, _should_plot


PI = np.pi


def plot_signal(
    x, y, show_zeros, show_segments, show_auc, highlight_time, highlight_auc,
    label, identifier, output_path
):
    """To plot y versus x.

    Args:
    ----
        x (array): the values for the x-axis.
        y (array): the values for the y-axis.
        show_zeros (bool): to highlight the zeros: the x points such as y(x) = 0.
        show_segments (bool): to distinguish between the positive and negative
                              parts of the curve.
        show_auc (bool): to distinguish between the positive and negative
                         areas of the curve.
        highlight_time (tuple): to highlight breathing cycles with a specific time.
        highlight_auc (tuple): to highlight breathing cycles with a specific area.
        label (str): the label of the curve.
        identifier (str): the identifier of the object whose signal is plotted.
        output_path (str): to choose where to save the figure, if applicable.

    Returns:
    -------
        None. Plot the figure.

    """
    positive_segments = features.get_segments(x, y)[0]
    negative_segments = features.get_segments(x, y)[1]

    positive_auc = features.get_auc_value(
        segments=positive_segments,
        return_mean=False,
        verbose=False,
        n_digits=2,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
    )

    negative_auc = features.get_auc_value(
        segments=negative_segments,
        return_mean=False,
        verbose=False,
        n_digits=2,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
    )

    pos_normalized = positive_auc / np.max(positive_auc)
    neg_normalized = negative_auc / np.min(negative_auc)

    cmap_pos, cmap_neg = plt.cm.GnBu, plt.cm.RdPu
    cmap_pos_normalized = cmap_pos(
        np.linspace(np.min(pos_normalized), np.max(pos_normalized), len(positive_auc))
    )
    cmap_neg_normalized = cmap_neg(
        np.linspace(np.max(neg_normalized), np.min(neg_normalized), len(negative_auc))
    )
    global_cmap = mcolors.ListedColormap(
        np.concatenate([cmap_neg_normalized, cmap_pos_normalized])
    )

    bbox_style = {
        "facecolor": "whitesmoke",
        "boxstyle": "round,pad=0.3",
        "edgecolor": "k",
        "lw": 0.2
    }

    fig, ax = plt.subplots(figsize=(14, 2))

    if show_segments:
        for i, (x_pos, y_pos) in enumerate(positive_segments):
            pos_label = "Air flow rate > 0" if i == 1 else ""
            ax.plot(x_pos, y_pos, label=pos_label, c="tab:blue")
        for i, (x_neg, y_neg) in enumerate(negative_segments):
            neg_label = "Air flow rate < 0" if i == 1 else ""
            ax.plot(x_neg, y_neg, label=neg_label, c="tab:orange")
    else:
        ax.plot(x, y, label=label, c="tab:gray", lw=1)
        ax.axhline(y=0, c="grey", linestyle=":", lw=1)

    if show_zeros:
        zeros = np.where(y == 0)[0]
        ax.scatter(x[zeros], y[zeros], zorder=2, c="gold", s=9, lw=0.4, edgecolor="k")

    if show_auc:
        for i, (xp, yp) in enumerate(positive_segments):
            color = cmap_pos(pos_normalized[i])
            ax.fill_between(xp, yp, color=color, alpha=1)

        for j, (xn, yn) in enumerate(negative_segments):
            color = cmap_neg(neg_normalized[j])
            ax.fill_between(xn, yn, color=color, alpha=1)

        sm = ScalarMappable(
            norm=mcolors.TwoSlopeNorm(
                vmin=np.min(negative_auc), vcenter=0, vmax=np.max(positive_auc)
            ),
            cmap=global_cmap,
        )
        sm.set_array([])
        cbar = fig.colorbar(sm, ax=ax, orientation="vertical")
        cbar.set_label("Area under the curve", labelpad=15)

    if highlight_time:
        for s in (positive_segments, negative_segments):
            cmap = cmap_pos if s == positive_segments else cmap_neg
            normalized = pos_normalized if s == positive_segments else neg_normalized
            for t in highlight_time:
                for i, (xs, ys) in enumerate(s):
                    if scientific_round((xs[-1] - xs[0]), n_digits=3) == t:
                        color = cmap(normalized[i])
                        ax.fill_between(xs, ys, color=color, alpha=1)
                        max_y = max(ys) if max(ys) > 0 else min(ys)
                        ax.text(
                            x=(xs[0] + xs[-1]) / 2,
                            y=max_y,
                            s=f"t={t}",
                            fontsize=8,
                            backgroundcolor="whitesmoke",
                            bbox=bbox_style
                        )

    if highlight_auc:
        for s in (positive_segments, negative_segments):
            cmap = cmap_pos if s == positive_segments else cmap_neg
            normalized = pos_normalized if s == positive_segments else neg_normalized
            for a in highlight_auc:
                for i, (xs, ys) in enumerate(s):
                    if scientific_round(trapezoid(y=ys, x=xs), n_digits=3) == a:
                        color = cmap(normalized[i])
                        ax.fill_between(xs, ys, color=color, alpha=1)
                        max_y = max(ys) if max(ys) > 0 else min(ys)
                        ax.text(
                            x=(xs[0] + xs[-1]) / 2,
                            y=max_y,
                            s=f"auc={a}",
                            fontsize=8,
                            backgroundcolor="whitesmoke",
                            bbox=bbox_style
                        )

    ax.set_xlabel("time (s)", labelpad=10)
    if label == "air flow rate":
        ax.set_ylabel("Air flow rate", labelpad=10)
        ax.set_title(
            f"{identifier}: {len(positive_segments)} positive segments & {len(negative_segments)} negative segments detected",
            fontsize=9,
            c="k",
            backgroundcolor="whitesmoke",
            bbox={
                "facecolor": "whitesmoke",
                "boxstyle": "round,pad=0.3",
                "edgecolor": "silver",
                "lw": 0.2,
            },
        )
    else:
        ax.set_ylabel("Amplitude", labelpad=10)

    ax.grid(alpha=0.8, linestyle=":", ms=0.1, zorder=1)
    ax.legend(fontsize=8, loc="upper left", bbox_to_anchor=(0, 1.2))
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")

    if _should_plot(func="data_merger") and _should_plot(func="get_overview"):
        plt.show()
    else:
        plt.close(fig)


def plot_peaks(x, y, identifier, which_peaks, distance, output_path):
    """
    Calibration of peaks detection
    = test of the distance that correctly detects all the peaks.

    Args:
    ----
        x (array): the values for the x-axis.
        y (array): the values for the y-axis.
        identifier (str): the identifier of the object whose peaks are plotted.
        which_peaks (str): to consider either top or bottom peaks.
        distance (int): the minimum distance between two neighbouring peaks.
        output_path (str): to choose where to save the figure, if applicable.

    Returns:
    -------
        None. Plots a control figure.

    """
    top_peaks, _ = find_peaks(y, distance=distance)
    bottom_peaks, _ = find_peaks(-y, distance=distance)

    fig, ax = plt.subplots(figsize=(14, 2))

    ax.plot(x, y, c="tab:blue", label="your signal", zorder=1)

    match which_peaks:
        case "top":
            ax.scatter(
                x[top_peaks],
                y[top_peaks],
                s=10,
                marker="x",
                lw=2,
                c="red",
                label=f"{len(top_peaks)} detected peaks (distance = {distance})",
                zorder=2,
            )
        case "bottom":
            ax.scatter(
                x[bottom_peaks],
                y[bottom_peaks],
                s=10,
                marker="x",
                lw=2,
                c="red",
                label=f"{len(bottom_peaks)} detected peaks (distance = {distance})",
                zorder=2,
            )
        case _:
            raise ValueError(
                "Argument 'which_peaks' must be either 'top' or 'bottom'. "
                f"Not '{which_peaks}'."
            )

    min_s, max_s = min(y), max(y)
    s_amp = max_s - min_s
    min_b, max_b = min_s - 0.25 * s_amp, max_s + 0.25 * s_amp
    ax.set_ylim(min_b, max_b)
    ax.set_xlabel("time (s)", labelpad=10)
    ax.set_ylabel("Air flow rate", labelpad=10)
    ax.grid(alpha=0.8, linestyle=":", ms=0.5)
    ax.legend(prop={"size": 8}, loc="upper left", bbox_to_anchor=(0, 1.25))
    ax.set_title(f"Detected peaks for: {identifier}")

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")


def plot_features_distribution(*args, identifier, stat, kde_min, output_path):
    """
    To get the distribution of each feature of the 'BreathingFlow' object.

    Args:
    ----
        *args (array): all the values of one of the signal features.
        identifier (str): the identifier of the object whose distribution is plotted.
        stat (str): aggregate statistic to compute in each bin.
        kde_min (bool): to highlight the minimum of the kernel density estimation.
        output_path (str): to choose where to save the figure, if applicable.

    Returns:
    -------
        None. Plots the distribution.

    """
    x1, x2, x3, x4 = args
    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(nrows=2, ncols=2, figsize=(12, 6))

    sns.histplot(data=x1, kde=True, stat=stat, ax=ax1)
    sns.histplot(data=x2, kde=True, stat=stat, ax=ax2)
    sns.histplot(data=x3, kde=True, stat=stat, ax=ax3)
    sns.histplot(data=x4, kde=True, stat=stat, ax=ax4)

    for ax in fig.axes:
        ax.grid(alpha=0.8, linestyle=":", ms=0.5)
        ax.lines[0].set_color("crimson")
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        if kde_min:
            kde = ax.lines[0]
            x_kde, y_kde = kde.get_xdata(), kde.get_ydata()
            min_index_kde = np.argmin(y_kde)
            min_x_kde, min_y_kde = x_kde[min_index_kde], y_kde[min_index_kde]

            ax.scatter(min_x_kde, min_y_kde, c="crimson", edgecolor="k", zorder=2)
            ax.annotate(
                f"min = {min_x_kde:.3f}",
                xy=(min_x_kde, min_y_kde),
                xytext=(min_x_kde, min_y_kde + 0.1),
                arrowprops=dict(
                    arrowstyle="->",
                    connectionstyle="arc3,rad=0.3",
                    shrinkB=5
                ),
                fontsize=9,
                bbox=dict(
                    boxstyle="round,pad=0.3",
                    edgecolor="black",
                    facecolor="whitesmoke",
                    lw=0.5,
                    alpha=0.8
                )
            )

    for ax, lab, x in zip(
        fig.axes,
        ("time (AUC > 0) (s)", "time (AUC < 0) (s)", "AUC > 0", "AUC < 0"),
        args,
    ):
        ax.set_xlabel(lab, labelpad=10)
        if lab.endswith("(s)"):
            lab = lab[:-3]

        match stat:
            case "probability":
                ax.set_ylabel(rf"$\mathbb{{P}}$ ({lab})", labelpad=10)
            case "count":
                ax.set_ylabel(rf"count ({lab})", labelpad=10)
            case "frequency":
                ax.set_ylabel(rf"frequency ({lab})", labelpad=10)
            case "percent":
                ax.set_ylabel(rf"percentage ({lab})", labelpad=10)
            case "density":
                ax.set_ylabel(rf"density ({lab})", labelpad=10)

        ax.set_title(f"Distribution of {lab} (n = {len(x)})", pad=10)

    fig.suptitle(f"Distribution of features of: {identifier}")

    fig.tight_layout()

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")

    if _should_plot(func="data_merger") and _should_plot(func="get_overview"):
        plt.show()
    else:
        plt.close(fig)


def plot_phase_portrait(x, y, identifier, time_delay, hz, color_scheme, output_path):
    """
    To plot the phase portrait of the time series y.

    Args:
    ----
        x (array): time axis of the time series.
        y (array): y axis of the time series.
        identifier (str): the identifier of the object whose phase portrait is plotted.
        time_delay (float): parameter for phase portrait offset: y(x) vs. y(x+t).
        hz (int): the sampling rate of the time series.
        color_scheme (str): whether the color is defined from time or respiratory phases.
        output_path (str): to choose where to save the figure, if applicable.

    Returns:
    -------
        None. Plots the phase portrait of the time series.

    """
    if color_scheme not in ("time", "phases"):
        raise ValueError(
            f"color_scheme should be either 'time' or 'phases'. Not {color_scheme}."
        )

    time_delay *= hz
    time_delay = int(time_delay)

    # 2D plot.
    y0 = y[:-time_delay]
    y_tau = y[time_delay:]

    # 3D plot.
    x_3d = y[:-3*time_delay:]
    y_3d = y[time_delay:-2*time_delay]
    z_3d = y[3*time_delay:]

    if color_scheme == "time":
        cmap = plt.cm.viridis(np.linspace(0, 1, len(y)))
    else:
        cmap = np.where(y > 0, "tab:blue", "tab:orange")

    fig = plt.figure(figsize=(14, 6))
    gs = fig.add_gridspec(2, 2, height_ratios=[1, 2], width_ratios=[1, 1.5])
    ax1 = fig.add_subplot(gs[0, :])
    ax2 = fig.add_subplot(gs[1, 0])
    ax3 = fig.add_subplot(gs[1, 1], projection="3d")

    # ax1.
    segments_1 = np.array([[[x[i], y[i]], [x[i+1], y[i+1]]] for i in range(len(x)-1)])
    lc1 = LineCollection(segments_1, colors=cmap)
    ax1.add_collection(lc1)
    legend_1 = Line2D([0], [0], color=cmap[0], label="air flow rate")
    ax1.legend(handles=[legend_1], fontsize=8, loc="upper left")
    ax1.autoscale()

    # ax2.
    segments_2 = np.array([[[y0[i], y_tau[i]], [y0[i+1], y_tau[i+1]]] for i in range(len(y0)-1)])
    lc2 = LineCollection(segments_2, colors=cmap)
    ax2.add_collection(lc2)
    legend_2 = Line2D([0], [0], color=cmap[0], label="phase portrait 2D")
    ax2.legend(handles=[legend_2], fontsize=8, loc="upper left")
    ax2.autoscale()

    segments_3d = np.array([[[x_3d[i], y_3d[i], z_3d[i]], [x_3d[i+1], y_3d[i+1], z_3d[i+1]]] for i in range(len(x_3d)-1)])
    lc3 = Line3DCollection(segments_3d, colors=cmap, linewidths=2)
    ax3.add_collection(lc3)
    legend_3 = Line2D([0], [0], color=cmap[0], label="phase portrait 3D")
    ax3.legend(handles=[legend_3], fontsize=8, loc="upper left")

    ax3.set_xlim(np.min(x_3d), np.max(x_3d))
    ax3.set_ylim(np.min(y_3d), np.max(y_3d))
    ax3.set_zlim(np.min(z_3d), np.max(z_3d))

    for ax in fig.get_axes():
        ax.grid(alpha=0.8, linestyle=":", ms=0.1, zorder=1)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

    ax1.set_xlabel("time (s)", labelpad=10)
    ax1.set_ylabel("Air flow rate", labelpad=10)

    ax2.set_xlabel("y(t)", labelpad=10)
    ax2.set_ylabel(f"y(t + {time_delay / hz})", labelpad=10)

    ax3.set_xlabel("y(t)", labelpad=10)
    ax3.set_ylabel(f"y(t + {time_delay / hz})", labelpad=10)
    ax3.set_zlabel(f"y(t + {2*time_delay / hz})", labelpad=10)

    fig.suptitle(f"Phase portrait of: {identifier}")

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")

    if _should_plot(func="data_merger") and _should_plot(func="get_overview"):
        plt.show()
    else:
        plt.close(fig)


def plot_movements(y1, y2, y3, overlay, output_path):
    """
    To plot air flow rate and breathing movements on the same plot.

    Args:
    ----
        y1 (BreathingFlow): Air flow rate.
        y2 (BreathingMovement): Thoracic movements.
        y3 (BreathingMovement): Abdominal movements.
        overlay (bool): whether or not to superimpose respiratory movements.
        output_path (str): to choose where to save the figure, if applicable.

    Returns:
    -------
        None. Plots the three (or two) curves.

    """
    if y1 is not None:
        y1 = {
            "x": y1.absolute_time,
            "y": y1.flow,
            "label": "air flow rate",
            "ylabel": "Air Flow Rate",
            "color": "tab:gray",
            "title": f"Air flow rate of: {y1.identifier}"
        }

    y2 = {
        "x": y2.absolute_time,
        "y": y2.movements,
        "label": y2.movement_type,
        "ylabel": "Amplitude",
        "color": "tab:red",
        "title": f"Movements of: {y2.identifier}"
    }

    y3 = {
        "x": y3.absolute_time,
        "y": y3.movements,
        "label": y3.movement_type,
        "ylabel": "Amplitude",
        "color": "tab:purple",
        "title": f"Movements of: {y3.identifier}"
    }

    if y1 is None and overlay:
        rows, plots = 1, [[y2, y3]]
    elif y1 is None:
        rows, plots = 2, [[y2], [y3]]
    elif overlay:
        rows, plots = 2, [[y1], [y2, y3]]
    else:
        rows, plots = 3, [[y1], [y2], [y3]]

    fig, axes = plt.subplots(figsize=(14, rows*2), nrows=rows)
    if rows == 1:
        axes = [axes]

    for ax, series in zip(axes, plots):
        for s in series:
            ax.plot(s["x"], s["y"], label=s["label"], color=s["color"])

        ax.set_title(
            series[0]["title"],
            fontsize=9,
            c="k",
            backgroundcolor="whitesmoke",
            bbox={
                "facecolor": "whitesmoke",
                "boxstyle": "round,pad=0.3",
                "edgecolor": "silver",
                "lw": 0.2,
                }
        )
        ax.grid(alpha=0.8, linestyle=":", ms=0.1, zorder=1)
        ax.set_ylabel(series[0]["ylabel"], labelpad=10)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.legend(fontsize=8, loc="upper left")

    axes[-1].set_xlabel("time (s)", labelpad=10)

    plt.tight_layout()

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")

    if _should_plot(func="data_merger") and _should_plot(func="get_overview"):
        plt.show()
    else:
        plt.close(fig)


def plot_phase_difference(
        time, y1, y2, segment_indices, phase_diff, output_path, view
):
    """
    To plot the phase difference (phase shift) between y1 et y2.

    Args:
    ----
        time (array): time vector.
        y1 (BreathingMovement): thoracic movements.
        y2 (BreathingMovement): abdominal movements.
        segment_indices (list): indices (time points) for which the phase shift is plotted.
        phase_diff (list): values of phase shifts for each segment_indices.
        output_path (str): to choose where to save the figure, if applicable.
        view (bool): whether or not to display the figure.

    Returns:
    -------
        None. Plots the figure.

    """
    y1 = {"y": y1.movements, "label": y1.movement_type, "id": y1.identifier}
    y2 = {"y": y2.movements, "label": y2.movement_type, "id": y2.identifier}

    fig = plt.figure(figsize=(14, 4.65))
    gs = fig.add_gridspec(2, 1, height_ratios=[1, 1.5])
    ax1 = fig.add_subplot(gs[0, :])
    ax2 = fig.add_subplot(gs[1, 0], sharex=ax1)

    ax1.plot(time, y1["y"], c="tab:red", label=y1["label"])
    ax1.plot(time, y2["y"], c="tab:purple", label=y2["label"])
    ax1.set_ylabel("Amplitude", labelpad=10)
    ax1.set_title(
        f"Movements of: {y1['id']}",
        fontsize=9,
        c="k",
        backgroundcolor="whitesmoke",
        bbox={
            "facecolor": "whitesmoke",
            "boxstyle": "round,pad=0.3",
            "edgecolor": "silver",
            "lw": 0.2,
            }
    )

    ax2.plot(
        segment_indices,
        phase_diff,
        label="phase difference",
        c="navy",
        marker="o",
        markeredgecolor="w",
        markeredgewidth=0.6,
        lw=0.75
    )
    ax2.set_ylabel("Phase difference (radians)", labelpad=10)
    ax2.set_xlabel("time (s)", labelpad=10)
    ax2.set_ylim(-3.6, 3.6)
    ax1.set_title(
        f"Phase shift in respiratory movements of: {y1['id']}",
        fontsize=9,
        c="k",
        backgroundcolor="whitesmoke",
        bbox={
            "facecolor": "whitesmoke",
            "boxstyle": "round,pad=0.3",
            "edgecolor": "silver",
            "lw": 0.2,
            }
    )

    ax2.axhspan(ymin=7*PI/6, ymax=5*PI/6, facecolor="#e31a1c", alpha=0.22)
    ax2.axhspan(ymin=-7*PI/6, ymax=-5*PI/6, facecolor="#e31a1c", alpha=0.22)

    ax2.axhspan(ymin=5*PI/6, ymax=PI/2, facecolor="#fd8d3c", alpha=0.22)
    ax2.axhspan(ymin=-5*PI/6, ymax=-PI/2, facecolor="#fd8d3c", alpha=0.22)

    ax2.axhspan(ymin=PI/2, ymax=PI/6, facecolor="#fed976", alpha=0.22)
    ax2.axhspan(ymin=-PI/2, ymax=-PI/6, facecolor="#fed976", alpha=0.22)

    ax2.axhspan(ymin=-PI/6, ymax=PI/6, facecolor="#78c679", alpha=0.22)

    for ax in (ax1, ax2):
        ax.grid(alpha=0.8, linestyle=":", ms=0.1, zorder=1)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.legend(fontsize=8, loc="upper left")

    plt.tight_layout()

    if view:
        plt.show()
    else:
        plt.close(fig)

    if output_path:
        fig.savefig(output_path, bbox_inches="tight")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Useful functions for managing secondary tasks.

Created on Thu Apr 10 09:48:33 2025
@author: CoustilletT
"""


from importlib.resources import files, as_file
from functools import wraps
from inspect import signature, stack
import os
from pathlib import Path
import re

import numpy as np
import pandas as pd
import pypdf


class ComparableMixin:
    def __eq__(self, other_object, top_level=True):
        """To determine whether two instances are equal.

        Equality is based on a comparison of all instance attributes.

        """
        if not isinstance(other_object, self.__class__):
            return NotImplemented

        diffs = []

        for attr in self.__dict__:
            attr_1 = getattr(self, attr)
            attr_2 = getattr(other_object, attr)

            if isinstance(attr_1, np.ndarray) and isinstance(attr_2, np.ndarray):
                if not np.array_equiv(attr_1, attr_2):
                    diffs.append(attr.lstrip("_"))

            elif isinstance(attr_1, ComparableMixin) and isinstance(attr_2, ComparableMixin):
                if not attr_1.__eq__(attr_2, top_level=False):
                    diffs.append(attr.lstrip("_"))
            else:
                if attr_1 != attr_2:
                    diffs.append(attr.lstrip("_"))

        if diffs and top_level:
            print(f"Different attributes: {', '.join(diffs)}.")
            return False
        return not diffs


# Function (decorator) that handle argument types for object methods.
def enforce_type_arg(**arg_types):
    """Decorate a function to enforce a method argument to be of certain type."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            bound_args = signature(func).bind(*args, **kwargs)
            bound_args.apply_defaults()

            for arg_name, expected_type in arg_types.items():
                if arg_name not in bound_args.arguments:
                    continue

                arg_type = bound_args.arguments[arg_name]
                if not isinstance(arg_type, expected_type):
                    raise TypeError(
                        f"'{arg_name}' argument must be of type "
                        f"'{expected_type.__name__}', not "
                        f"'{type(arg_type).__name__}'."
                    )

            return func(*args, **kwargs)
        return wrapper
    return decorator


def scientific_round(x, n_digits):
    """
    To evenly round x to the given number significant digits.

    Args:
    ----
        x (tuple, float): a float or a tuple of floats.
        n_digits (int): number of significant digits.

    Returns:
    -------
        tuple or float: rounded tuple or float.

    """
    if isinstance(x, (tuple, list, np.ndarray)):
        return tuple(float(f"{element:.{n_digits - 1}e}") for element in x)
    if isinstance(x, float):
        return float(f"{x:.{n_digits - 1}e}")


def data_merger(*args, table_name, output_directory=None):
    """
    To combine data from several respiratory signals into an unique summary table.

    Args:
    ----
        *args (BreathingFlow/Signals): Respiratory air flow rates or movements
                                       instantiated and analysed using
                                       'BreathingFlow' or 'BreathingSignals' objects.
        table_name (str): name of the summary table.
        output_directory (str, optional): backup directory for the summary table.
                                          Defaults to None.

    Raises:
    ------
        TypeError: forces all arguments to be of the same type
                  ('BreathingFlow' or 'BreathingSignals').

    Returns:
    -------
        merged_df (pandas.DataFrame): merged data; 1 row = features for a file.

    """
    from .breathingflow import BreathingFlow
    from .breathingsignals import BreathingSignals

    if not args:
        return pd.DataFrame()

    class_to_merge = type(args[0])

    if not all(isinstance(arg, class_to_merge) for arg in args):
        raise TypeError(f"All arguments must be of the same type.")

    if class_to_merge.__name__ not in {"BreathingFlow", "BreathingSignals"}:
        raise TypeError(f"Unsupported class for merging: {class_to_merge.__name__}")

    overview_1f = [arg.get_overview() for arg in args]
    merged_df = pd.concat(overview_1f)

    info_df = [arg.get_info(shape="df") for arg in args]
    merged_info = pd.concat(info_df)

    if output_directory:
        backup_dir = os.path.join(output_directory, table_name)
        os.makedirs(backup_dir, exist_ok=True)
        output_path = os.path.join(backup_dir, f"overview_{table_name}")

        # Table.
        with pd.ExcelWriter(f"{output_path}.xlsx", engine="xlsxwriter") as w:
            if len(table_name) > 26:
                table_name = table_name[:26]
            merged_df.to_excel(w, sheet_name=f"data_{table_name}")
            merged_info.to_excel(w, sheet_name=f"info_{table_name}", index=False)

        # Figures.
        fig_dir = os.path.join(backup_dir, "figures")
        os.makedirs(fig_dir, exist_ok=True)

        flow_dir = os.path.join(fig_dir, "flow")
        distrib_dir = os.path.join(fig_dir, "feat_distrib")
        portrait_dir = os.path.join(fig_dir, "phase_portrait")
        mov_dir = os.path.join(fig_dir, "movements")

        if isinstance(args[0], BreathingFlow) or getattr(args[0], "flow", None) is not None:
            os.makedirs(flow_dir, exist_ok=True)
            os.makedirs(distrib_dir, exist_ok=True)
            os.makedirs(portrait_dir, exist_ok=True)

            for arg in args:
                if isinstance(arg, BreathingSignals):
                    arg = arg.flow
                identifier = re.sub(r"[^a-zA-Z0-9\s]", "", arg.identifier)
                ext = f"_{identifier}.pdf"
                arg.plot(
                    show_auc=True,
                    output_path=os.path.join(flow_dir, f"flow{ext}")
                )

                arg.plot_distribution(
                    output_path=os.path.join(distrib_dir, f"features{ext}")
                )

                arg.plot_phase_portrait(
                    color_scheme="phases",
                    output_path=os.path.join(portrait_dir, f"phase_portrait{ext}"))

        if isinstance(args[0], BreathingSignals):
            os.makedirs(mov_dir, exist_ok=True)

            for arg in args:
                identifier = re.sub(r"[^a-zA-Z0-9\s]", "", arg.identifier)
                ext = f"_{identifier}.pdf"
                arg.plot(output_path=os.path.join(mov_dir, f"movements{ext}"))
                arg.get_coherence(
                    view=False,
                    output_path=os.path.join(mov_dir, f"coherence{ext}")
                )

        # Merge PDF.
        for d in Path(fig_dir).iterdir():
            if d.is_dir():
                if d == Path(mov_dir):
                    for t in ["coherence", "movements"]:
                        files = list(Path(mov_dir).glob(f"{t}_*.pdf"))
                        if files:
                            pdf_merger(
                                *files,
                                output_path=os.path.join(os.path.dirname(fig_dir), f"{t}_{table_name}.pdf")
                            )
                else:
                    figs = list(d.glob("*.pdf"))
                    if figs:
                        pdf_merger(
                            *figs,
                            output_path=os.path.join(os.path.dirname(fig_dir), f"{d.name}_{table_name}.pdf")
                        )

    return merged_df, merged_info


def print_source():
    """To print the origin of breathing-like signals."""
    source = files("pybreathe.datasets").joinpath("readme.txt")
    with as_file(source) as f:
        blocs = f.read_text(encoding="utf-8").split("info:")[1:]
        return {
            b.split("\n", 1)[0]: b.split("\n", 1)[1].rstrip() for b in blocs
        }


def _check_type(value, expected_type, name, allow_none=False):
    """To ensure that the type passed to the class constructor is correct."""
    if value is None and allow_none:
        return
    if not isinstance(value, expected_type):
        raise TypeError(
            f"Expected '{name}' to be of type '{expected_type.__name__}', "
            f"got {type(value).__name__}."
        )


def create_absolute_time(time_vector, hz):
    """
    To create an absolute time vector starting from 0.

    Args:
    ----
        time_vector (array): any time vector.
        hz (int): the sampling rate of the time vector.

    Returns:
    -------
        array: time vector in an absolute format: [0.000, 0.002, 0.004, ...].

    """
    return np.linspace(
        0, len(time_vector) / hz, len(time_vector), endpoint=False
    )


def to_dataframe(identifier, overview_dict):
    """
    To convert a dictionary into a specially formatted Dataframe.

    Args:
    ----
        identifier (str): data dictionary identifier.
        overview_dict (dict): dictionary hosting all the signal features.

    Returns:
    -------
        multicols_df (pandas.DataFrame): dataframe summarising the
                                         features (freq, auc, times).

    """
    data_tuples = [
        ((key, sub_key), value)
        for key, sub_dict in overview_dict.items()
        for sub_key, value in sub_dict.items()
    ]
    multicols_df = pd.DataFrame.from_dict(dict(data_tuples), orient="index").T
    multicols_df.columns = pd.MultiIndex.from_tuples(multicols_df.columns)
    multicols_df.index.name = "identifier"
    multicols_df.index = [identifier]

    return multicols_df


def _should_plot(func):
    """Do not plot the figures when func() is called: return False."""
    return all(frame.function != func for frame in stack()[1:])


def pdf_merger(*args, output_path):
    """
    Merge multiple PDF files into a single PDF.

    Args:
    ----
        *args (str): file paths to PDF document to merge.
        output_path (str): file path where the merged PDF will be saved.

    Returns:
    -------
        None.

    """
    merger = pypdf.PdfWriter()

    for pdf in args:
        merger.append(pdf)

    merger.write(output_path)
    merger.close()

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions required to extract signal features.

Created on Thu Apr  3 10:10:59 2025
@author: thibaut
"""


import numpy as np
from numpy import ma
import pandas as pd
from scipy.integrate import trapezoid
from scipy.signal import find_peaks, periodogram, welch, windows, argrelextrema

from .utils import scientific_round


def compute_sampling_rate(x):
    """
    To get the sampling rate of a time vector x.

    Args:
    ----
        x (array): a time vector in seconds format ([0.001, 0.002, 0.003, ..]).

    Returns:
    -------
        int: the sampling rate in Hz (s-1) = the number of points in 1 s.

    """
    time_delta = pd.to_timedelta(x, unit="s")
    diff = time_delta.diff().value_counts().index.tolist()[0]

    return int(pd.Timedelta(seconds=1) / diff)


def zero_interpolation(x, y):
    """
    To find the 'True' zeros of a discretized signal.

    Args:
    ----
        x (array): a discretized vector of floats.
        y (array): a discretized vector y such that y = f(x).

    Returns:
    -------
        x_truncated (array): the input vector x to which have been added
                             all the x0 such that f(x0) = 0.
        y_truncated (array): the input vector y to which have been added
                             all the true zeros.

    Note:
    ----
        the 'y' vector must contains positive and negative values.

    """

    if np.all(np.sign(y) == np.sign(y[0])):
        raise ValueError(
            "Single-sign air flow rate detected. Please use 'detrend_y=True' "
            "or provide a valid air flow rate.")

    crossing_indices = np.where(
        np.array([0 if e in (-1, 1) else e for e in np.diff(np.sign(y))])
    )[0]

    x_zeros, y_zeros = [], []
    for i in crossing_indices:
        x_upstream, x_downstream = x[i], x[i + 1]
        y_upstream, y_downstream = y[i], y[i + 1]

        zero = x_upstream - y_upstream * (x_downstream - x_upstream) / (
            y_downstream - y_upstream
        )
        x_zeros.append(zero)
        y_zeros.append(0.0)

    x_extended = np.concatenate((x, x_zeros))
    y_extended = np.concatenate((y, y_zeros))

    sorted_indices = np.argsort(x_extended)
    x_interpolated = x_extended[sorted_indices]
    y_interpolated = y_extended[sorted_indices]

    # Truncation to start and end with a 'True' zero.
    first_zero = np.where(y_interpolated == 0)[0][0]
    lasy_zero = np.where(y_interpolated == 0)[0][-1] + 1

    x_truncated = x_interpolated[first_zero:lasy_zero]
    y_truncated = y_interpolated[first_zero:lasy_zero]

    return x_truncated, y_truncated


def get_segments(x, y):
    """
    To get positive and negative segments of a signal.

    Args:
    ----
        x (array): a discretised vector of floats.
        y (array): a discretized vector y such that y = f(x).

    Returns:
    -------
        positive_segments (array): pairs (x, y) for which y is positive.
        negative_segments (array): pairs (x, y) for which y in negative.

    """
    positive_segments, negative_segments = [], []
    pos_x, pos_y, neg_x, neg_y = [], [], [], []

    for i, yi in enumerate(y):
        if yi > 0:
            if len(neg_x) > 0:
                negative_segments.append((neg_x, neg_y))
                neg_x, neg_y = [], []

            pos_x.append(x[i])
            pos_y.append(yi)

        elif yi < 0:
            if len(pos_x) > 0:
                positive_segments.append((pos_x, pos_y))
                pos_x, pos_y = [], []

            neg_x.append(x[i])
            neg_y.append(yi)

        else:
            if len(pos_x) > 0:
                positive_segments.append((pos_x, pos_y))
                pos_x, pos_y = [], []

            if len(neg_x) > 0:
                negative_segments.append((neg_x, neg_y))
                neg_x, neg_y = [], []

    if len(pos_x) > 0:
        positive_segments.append((pos_x, pos_y))

    if len(neg_x) > 0:
        negative_segments.append((neg_x, neg_y))

    def add_zeros(x, y, segments):
        """
        To make the positive and negative segments begin and end with 0.

        Args:
        ----
            x (array): a discretised vector of floats.
            y (array): a discretized vector y such that y = f(x).
            segments (list): list of arrays where the values of y
                             have always the same sign.

        Returns:
        -------
            segments (array): pairs (x, y) where y begins and ends with a 0.

        """
        for s in segments:
            first_point = s[0][0]
            last_point = s[0][-1]

            first_index_zero = np.where(x == first_point)[0][0] - 1
            last_index_zero = np.where(x == last_point)[0][0]

            last_index_zero += 1 if (last_index_zero != (len(y) - 1)) else 0

            first_xzero = x[first_index_zero]
            first_yzero = y[first_index_zero]

            last_xzero = x[last_index_zero]
            last_yzero = y[last_index_zero]

            s[0].insert(0, first_xzero)
            s[0].append(last_xzero)
            s[1].insert(0, first_yzero)  # supposed to be 0.
            s[1].append(last_yzero)  # supposed to be 0.

        return segments

    positive_segments = add_zeros(x, y, positive_segments)
    negative_segments = add_zeros(x, y, negative_segments)

    return positive_segments, negative_segments


def get_segment_axis(segments, axis):
    """
    To get only the x or y values of the positive or negative segments.

    Args:
    ----
        segments (list): pairs (x, y) for which y is a single sign (> 0 or < 0).
        axis (str): to get only the x values or the y values (or both).

    Returns:
    -------
        list: list of values (x or y) of positive or negative segments.

    """

    match axis:
        case "x":
            return [x for x, _ in segments]
        case "y":
            return [y for _, y in segments]
        case "both":
            return segments
        case _:
            raise ValueError(
                "axis should be either 'x, 'y' or 'both'. "
                f"Not {axis}."
            )


def get_peaks(x, y, which_peaks, distance):
    """
    To get the top or bottom peaks of a signal.

    Args:
    ----
        x (array): a discretised vector of floats.
        y (array): a discretized vector y such that y = f(x).
        which_peaks (str): To get the top or bottom peaks.
                           Should be "top" or "bottom".
        distance (int): distance between two neighboring peaks.

    Returns:
    -------
        array: values of x such that y[x] is a peak.

    """
    if not (isinstance(distance, int) and distance > 0):
        raise ValueError(
            "To get top or bottom peaks, distance should be a "
            f"positive integer. Not '{distance}'. "
            "Please use 'test_distance method to set the right distance."
        )
    top_peaks, _ = find_peaks(y, distance=distance)
    bottom_peaks, _ = find_peaks(-y, distance=distance)

    if which_peaks == "top":
        return x[top_peaks]
    if which_peaks == "bottom":
        return x[bottom_peaks]
    return None


def frequency(signal, sampling_rate, method, which_peaks, distance, n_digits):
    """Get the frequency of a given signal.

    Args:
    ----
        signal (array): signal to test (list of numbers = discretized signal).
        sampling_rate (int): the sampling rate of the discretized signal.
        method (str): method to commpute frequency.
        which_peaks (str): if the method is 'peaks', which peaks should be
                           considered (top or bottom) ?
        distance (int): the minimum distance between two neighbouring peaks.
        n_digits (int): to round freq to n_digits significant digits.

    Returns:
    -------
        dominant_freq (float): breathing frequency (in respirations per min.)

    """
    match method:
        case "welch":
            fft_length = len(signal) // 3
            window = windows.hamming(fft_length)
            f, Pxx_den = welch(
                x=signal, fs=sampling_rate, window=window, nperseg=fft_length
            )

            dominant_freq = f[np.argmax(Pxx_den)]

        case "peaks":
            if not (isinstance(distance, int) and distance > 0):
                raise ValueError(
                    "To use the peak method, distance should be a "
                    f"positive integer. Not '{distance}'."
                    "Please use 'test_distance' method to set the right distance."
                )

            if which_peaks == "top":
                peaks, _ = find_peaks(x=signal, distance=distance)
            elif which_peaks == "bottom":
                peaks, _ = find_peaks(x=-signal, distance=distance)
            else:
                raise ValueError(
                    "Argument 'which_peaks' must be either 'top' or 'bottom'. "
                    f"Not '{which_peaks}'."
                )

            dominant_freq = sampling_rate / np.mean(np.diff(peaks))

        case "periodogram":
            window = windows.hamming(len(signal))
            f, Pxx_den = periodogram(x=signal, fs=sampling_rate, window=window)
            dominant_freq = f[np.argmax(Pxx_den)]

        case "fft":
            data_len = len(signal)
            sp = np.fft.fft(signal)
            freq = np.fft.fftfreq(data_len, d=1 / sampling_rate)
            dominant_freq = abs(freq[np.argmax(np.abs(sp[: data_len // 2]))])
        case _:
            raise ValueError(
                "'method' should be either 'welch', 'peaks', 'periodogram' or "
                f"'fft'. Not '{method}'."
            )

    # The frequency is in rpm.s-1; we want it in min.-1.
    dominant_freq *= 60

    return scientific_round(dominant_freq, n_digits=n_digits)


def get_auc_time(
    segments, return_mean, verbose, n_digits, lower_threshold, upper_threshold
):
    """
    To get the mean duration of segments when AUC is positive or negative.

    Args:
    ----
        segments (tuple): segments ([x0, x1, ...], [y0, y1, ...], ...) where
                          [y0, y1, ...] contains only values of the same sign.
        return_mean (bool): to return all values or only the mean.
        verbose (bool): to print (or not) results in human readable format.
        n_digits (int): to round time to n_digits significant digits.
        lower_threshold (float): to ignore values below the threshold.
        upper_threshold (float): to ignore values above the threshold.

    Returns:
    -------
        int: the mean duration of AUC
             (or array: all durations if return_mean = False).

    Note:
    ----
        To get segments, please use the 'get_segments' function.

    """
    points_of_interest = [s[0] for s in segments]
    all_durations = np.array([(p[-1] - p[0]) for p in points_of_interest])

    duration = ma.masked_outside(
        all_durations, lower_threshold, upper_threshold
    ).compressed()

    if return_mean:
        mean_duration = scientific_round(np.mean(duration), n_digits=n_digits)
        std_duration = scientific_round(np.std(duration), n_digits=n_digits)
        n_duration = len(duration)
        variability = scientific_round(std_duration/mean_duration*100, n_digits=n_digits)

        if verbose:
            print(f"mean = {mean_duration} ± {std_duration} "
                  f"(n = {n_duration}; var = {variability} %).")
        return mean_duration, std_duration, n_duration, variability

    return scientific_round(duration, n_digits=n_digits)


def get_auc_value(
    segments, return_mean, verbose, n_digits, lower_threshold, upper_threshold
):
    """
    To get the mean AUC of segments when AUC is positive or negative.

    Args:
    ----
        segments (tuple): segments ([x0, x1, ...], [y0, y1, ...], ...) where
                          [y0, y1, ...] contains only values of the same sign.
        return_mean (bool): to return all values or only the mean.
        verbose (bool) : to print (or not) results in human readable format.
        n_digits (int): to round auc value to n_digits significant digits.
        lower_threshold (float): to ignore values below the threshold.
        upper_threshold (float): to ignore values above the threshold.

    Returns:
    -------
        int: the mean of AUC (or array: all AUC if return_mean = False).

    Note:
    ----
        To get segments, please use the 'get_segments' function.

    """
    aucs = np.array([trapezoid(y=y, x=x) for x, y in segments])
    auc = ma.masked_outside(aucs, lower_threshold, upper_threshold).compressed()

    if return_mean:
        mean_auc = scientific_round(np.mean(auc), n_digits=n_digits)
        std_auc = scientific_round(np.std(auc), n_digits=n_digits)
        n_auc = len(auc)
        variability = scientific_round(std_auc/abs(mean_auc)*100, n_digits=n_digits)

        if verbose:
            print(f"mean = {mean_auc} ± {std_auc} (n = {n_auc}; var = {variability} %).")
        return mean_auc, std_auc, n_auc, variability

    return scientific_round(auc, n_digits=n_digits)


def extract_local_minima(y):
    """To extract the first and last local minimum of an array."""
    minima = argrelextrema(y, np.less)[0]
    first_min = minima[0]
    last_min = minima[-1] + 1

    return first_min, last_min



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script handling 'BreathingFlow' object: a discretized air flow rate.

Created on Wed Apr  2 08:40:50 2025
@author: CoustilletT
"""


import os

import numpy as np
import pandas as pd
from scipy.signal import detrend

from .instantiationmethods import (_from_file, _from_dataframe, _load_sinus,
                                   _load_breathing_like_signal_01,
                                   _load_breathing_like_signal_02,
                                   _process_time)
from .utils import (ComparableMixin, enforce_type_arg, scientific_round,
                    create_absolute_time, to_dataframe)
from . import featureextraction as features
from . import visualization


class BreathingFlow(ComparableMixin):
    """Breathing Air Flow rate."""

    @enforce_type_arg(identifier=str, detrend_y=bool)
    def __init__(self, identifier, raw_time, raw_flow, detrend_y):
        """
        Class constructor.

        Args:
        ----
            identifier (str): breathing signal identifier.
            raw_time (array): discretized time vector.
            raw_flow (array): discretized breathing signal (air flow rate).
            detrend_y (bool): to set the mean of the air flow rate at 0.

        Returns:
        -------
            None.

        """
        self.identifier = identifier
        self.raw_time = raw_time
        self.raw_flow = raw_flow

        self.processed_time = _process_time(raw_time)
        self.raw_absolute_time = create_absolute_time(self.raw_time, self.get_hz())

        if detrend_y:
            self.detrended_flow = detrend(self.raw_flow, type="constant")
            self.detrended_flow[np.isclose(self.detrended_flow, 0, atol=1e-12)] = 0

        y_to_be_interpolated = getattr(self, "detrended_flow", self.raw_flow)

        self.time, self.flow = features.zero_interpolation(
            x=self.processed_time, y=y_to_be_interpolated
        )

        self.absolute_time = create_absolute_time(self.time, self.get_hz())

        self._distance = None

        # Features
        self._frequency = None
        self._positive_time = None
        self._negative_time = None
        self._positive_auc = None
        self._negative_auc = None
        self._positive_minute_ventilation = None
        self._negative_minute_ventilation = None

    from_file = classmethod(_from_file)
    from_dataframe = classmethod(_from_dataframe)
    load_sinus = classmethod(_load_sinus)
    load_breathing_like_signal_01 = classmethod(_load_breathing_like_signal_01)
    load_breathing_like_signal_02 = classmethod(_load_breathing_like_signal_02)

    def __getitem__(self, key):
        """
        To allow a 'BreathingFlow' object to be sliced and used as a new object.

        Args:
        ----
            key (list): slice of shape [start:stop:steps].

        Returns:
        -------
            A new 'BreathingFlow' sliced object.

        """
        sliced_object = self.__class__(
            identifier=self.identifier,
            raw_time=self.time[key],
            raw_flow=self.flow[key],
            detrend_y=False,
        )

        if hasattr(self, "distance"):
            sliced_object._distance = self._distance

        return sliced_object

    def __len__(self):
        """To get the length of the signal (in number of points)."""
        return len(self.flow)

    @property
    def distance(self):
        """Getter."""
        return self._distance

    @property
    def frequency(self):
        """Getter."""
        return self._frequency

    @property
    def positive_time(self):
        """Getter."""
        return self._positive_time

    @property
    def negative_time(self):
        """Getter."""
        return self._negative_time

    @property
    def positive_auc(self):
        """Getter."""
        return self._positive_auc

    @property
    def negative_auc(self):
        """Getter."""
        return self._negative_auc

    @property
    def positive_minute_ventilation(self):
        """Getter."""
        return self._positive_minute_ventilation

    @property
    def negative_minute_ventilation(self):
        """Getter."""
        return self._negative_minute_ventilation

    def get_hz(self):
        """To get the sampling rate of the discretized breathing signal."""
        return features.compute_sampling_rate(x=self.processed_time)

    @enforce_type_arg(shape=str)
    def get_info(self, shape="dict"):
        """To get signal information (identifier, start, end and duration.)"""
        info = {
            "identifier": self.identifier,
            "start": self.raw_time[0],
            "end": self.raw_time[-1],
            "duration": str(pd.to_timedelta(
                self.raw_absolute_time[-1], unit="s")
            ).split()[-1]

        }

        if shape == "dict":
            return info
        if shape == "df":
            return pd.DataFrame([info])
        raise ValueError(f"shape must be 'dict' or 'df', not '{shape}'.")

    @enforce_type_arg(
        y=str,
        show_zeros=bool,
        show_segments=bool,
        show_auc=bool,
        highlight_time=tuple,
        highlight_auc=tuple,
        output_path=str,
    )
    def plot(
        self,
        y="flow",
        show_zeros=True,
        show_segments=False,
        show_auc=False,
        highlight_time=(),
        highlight_auc=(),
        output_path="",
    ):
        """
        To plot the air flow rate.

        Args:
        ----
            y (str, optional): the values for the y-axis.
            show_zeros (bool, optional): to highlight the zeros: the x points such as y(x) = 0.
                                          Defaults to True.
            show_segments (bool, optional): to distinguish between the positive and negative
                                            parts of the curve. Defaults to False.
            show_auc (bool, optional): to distinguish between the positive and negative
                                       areas of the curve. Defaults to False.
            highlight_time (tuple, optional): to highlight breathing cycles with a specific time.
            highlight_auc (tuple, optional): to highlight breathing cycles with a specific area.
            label (str): the label of the curve.
            output_path (str, optional): to choose where to save the figure, if applicable.
                                         Defaults to "" (figure not saved).

        Returns:
        -------
            None. Plot the figure.

        """
        match y:
            case "flow":
                x, y = self.time, self.flow
            case "raw_flow":
                x, y = self.raw_absolute_time, self.raw_flow
            case "detrended_flow":
                x, y = self.absolute_time, self.detrended_flow
            case _:
                raise AttributeError(
                    f"{self.__class__.__name__} object has no attribute '{y}'"
                )

        visualization.plot_signal(
            x=x,
            y=y,
            show_zeros=show_zeros,
            show_segments=show_segments,
            show_auc=show_auc,
            highlight_time=highlight_time,
            highlight_auc=highlight_auc,
            label="air flow rate",
            identifier=self.identifier,
            output_path=output_path,
        )

    @enforce_type_arg(axis=str)
    def get_positive_segments(self, axis="both"):
        """To get the pairs (x,y) for which the air flow rate is positive."""
        return features.get_segment_axis(
            segments=features.get_segments(self.time, self.flow)[0],
            axis=axis
        )

    @enforce_type_arg(axis=str)
    def get_negative_segments(self, axis="both"):
        """To get the pairs (x,y) for which the air flow rate is negative."""
        return features.get_segment_axis(
            segments=features.get_segments(self.time, self.flow)[1],
            axis=axis
        )

    @enforce_type_arg(which_peaks=str, distance=int, set_dist=bool, output_path=str)
    def test_distance(
        self, which_peaks="top", distance=0, set_dist=False, output_path=""
    ):
        """
        Calibration of peaks detection
        = test which distance should be assigned to the 'distance' attribute.

        Args:
        ----
            which_peaks (str, optional): to consider either top or bottom peaks. Defaults to 'top'.
            distance (int): the minimum distance between two neighbouring peaks.
            set_dist (bool, optionnal): to set the distance or not. Defaults to False.
            output_path (str, optional): to choose where to save the figure, if applicable.
                                         Defaults to "" (figure not saved).

        Returns:
        -------
            None. Plots a control figure.

        Note:
        ----
            You will probably have to test several distance values
            to find the one that detects all the peaks.

        """
        visualization.plot_peaks(
            x=self.time,
            y=self.flow,
            identifier=self.identifier,
            which_peaks=which_peaks,
            distance=distance,
            output_path=output_path,
        )

        if set_dist:
            self._distance = distance

    def get_top_peaks(self):
        """To get the top peaks of the air flow rate."""
        return features.get_peaks(
            x=self.time, y=self.flow, which_peaks="top", distance=self.distance
        )

    def get_bottom_peaks(self):
        """To get the bottom peaks of the air flow rate."""
        return features.get_peaks(
            x=self.time, y=self.flow, which_peaks="bottom", distance=self.distance
        )

    @enforce_type_arg(method=str, n_digits=int, save=bool)
    def get_frequency(self, method="welch", which_peaks=None, n_digits=3, save=False):
        """
        Get breathing frequency of the air flow rate (in respirations.min-1).

        Args:
        ----
            method (str, optional): method used to calculate the frequency.
                                    Defaults to "welch".
                                    Possible choices: 'welch', 'peaks', 'periodogram' or 'fft'.
            which_peaks (str, optional): if the method is 'peaks', which peaks should be considered?.
                                         Defaults to None.
            n_digits (int, optional): to round freq to n_digits significant
                                      digits. Defaults to 3.
            save (bool, optional): save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Returns:
        -------
            freq (float): breathing frequency in rpm.min-1.

        """
        freq = features.frequency(
            signal=getattr(self, "detrended_flow", self.raw_flow),
            sampling_rate=self.get_hz(),
            method=method,
            which_peaks=which_peaks,
            distance=self.distance,
            n_digits=n_digits,
        )

        if save:
            self._frequency = freq

        return freq

    @enforce_type_arg(
        return_mean=bool,
        verbose=bool,
        n_digits=int,
        lower_threshold=float,
        upper_threshold=float,
        save=bool,
    )
    def get_positive_time(
        self,
        return_mean=True,
        verbose=True,
        n_digits=3,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
        save=False,
    ):
        """To get the mean duration of positive segments (when AUC > 0).

        Args:
        ----
            return_mean (bool, optional): to return all values or only the mean.
                                          Defaults to True (= the mean).
            verbose (bool, optional): to print (or not) results in human
                                      readable format. Defaults to True.
            n_digits (int, optional): to round time to n_digits significant
                                      digits. Defaults to 3.
            lower_threshold (float, optional): to ignore values below the threshold.
                                         Defaults to - ∞.
            upper_threshold (float, optional): to ignore values above the threshold.
                                         Defaults to + ∞.
            save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Returns:
        -------
            positive_time: mean duration of positive segments (when AUC > 0)
                           (or all durations if return_mean = False).

        Note:
        ----
            AUC = Area Under the Curve.

        """
        pos_auc_time = features.get_auc_time(
            segments=self.get_positive_segments(),
            return_mean=return_mean,
            verbose=verbose,
            n_digits=n_digits,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
        )

        if save:
            self._positive_time = pos_auc_time

        return pos_auc_time

    @enforce_type_arg(
        return_mean=bool,
        verbose=bool,
        n_digits=int,
        lower_threshold=float,
        upper_threshold=float,
        save=bool,
    )
    def get_negative_time(
        self,
        return_mean=True,
        verbose=True,
        n_digits=3,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
        save=False,
    ):
        """
        To get the mean duration of negative segments (when AUC < 0).

        Args:
        ----
            return_mean (bool, optional): to return all values or only the mean.
                                          Defaults to True (= the mean).
            verbose (bool, optional): to print (or not) results in human
                                      readable format. Defaults to True.
            n_digits (int, optional): to round time to n_digits significant
                                      digits. Defaults to 3.
            lower_threshold (float, optional): to ignore values below the threshold.
                                         Defaults to - ∞.
            upper_threshold (float, optional): to ignore values above the threshold.
                                         Defaults to + ∞.
            save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Returns:
        -------
            negative_time: mean duration of negative segments (when AUC < 0).
                           (or all durations if return_mean = False).

        Note:
        ----
            AUC = Area Under the Curve.

        """
        neg_auc_time = features.get_auc_time(
            segments=self.get_negative_segments(),
            return_mean=return_mean,
            verbose=verbose,
            n_digits=n_digits,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
        )

        if save:
            self._negative_time = neg_auc_time

        return neg_auc_time

    @enforce_type_arg(
        return_mean=bool,
        verbose=bool,
        n_digits=int,
        lower_threshold=float,
        upper_threshold=float,
        save=bool,
    )
    def get_positive_auc(
        self,
        return_mean=True,
        verbose=True,
        n_digits=3,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
        save=False,
    ):
        """
        To get the mean AUC of positive segments (when AUC > 0).

        Args:
        ----
            return_mean (bool, optional): to return all values or only the mean.
                                          Defaults to True (= the mean).
            verbose (bool, optional): to print (or not) results in human
                                      readable format. Defaults to True.
            n_digits (int, optional): to round auc value to n_digits significant
                                      digits. Defaults to 3.
            lower_threshold (float, optional): to ignore values below the threshold.
                                         Defaults to - ∞.
            upper_threshold (float, optional): to ignore values above the threshold.
                                         Defaults to + ∞.
            save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Returns:
        -------
            positive_auc: mean AUC of positive segments (when AUC > 0).
                          (or each AUC of each segment if return_mean = False).

        Note:
        ----
            AUC = Area Under the Curve.

        """
        pos_auc_val = features.get_auc_value(
            segments=self.get_positive_segments(),
            return_mean=return_mean,
            verbose=verbose,
            n_digits=n_digits,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
        )

        if save:
            self._positive_auc = pos_auc_val

        return pos_auc_val

    @enforce_type_arg(
        return_mean=bool,
        verbose=bool,
        n_digits=int,
        lower_threshold=float,
        upper_threshold=float,
        save=bool,
    )
    def get_negative_auc(
        self,
        return_mean=True,
        verbose=True,
        n_digits=3,
        lower_threshold=-np.inf,
        upper_threshold=np.inf,
        save=False,
    ):
        """
        To get the mean AUC of negative segments (when AUC < 0).

        Args:
        ----
            return_mean (bool, optional): to return all values or only the mean.
                                          Defaults to True (= the mean).
            verbose (bool, optional): to print (or not) results in human
                                      readable format. Defaults to True.
            n_digits (int, optional): to round auc value to n_digits significant
                                      digits. Defaults to 3.
            lower_threshold (float, optional): to ignore values below the threshold.
                                         Defaults to - ∞.
            upper_threshold (float, optional): to ignore values above the threshold.
                                         Defaults to + ∞.
            save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Returns:
        -------
            negative_auc: mean AUC of negative segments (when AUC < 0).
                          (or each AUC of each segment if return_mean = False).

        Note:
        ----
            AUC = Area Under the Curve.
            'threshold' must be negative as AUC is also negative, e.g.,
            if AUC values = [- 0.2, - 0.18, - 0.23, - 0.01], then to remove the
            - 0.01 value, threshold should be - 0.05 for example.

        """
        neg_auc_val = features.get_auc_value(
            segments=self.get_negative_segments(),
            return_mean=return_mean,
            verbose=verbose,
            n_digits=n_digits,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
        )

        if save:
            self._negative_auc = neg_auc_val

        return neg_auc_val

    @enforce_type_arg(verbose=bool, n_digits=int, save=bool)
    def get_minute_ventilation(self, verbose=True, n_digits=3, save=False):
        """
        To get minute ventilation of positive and negative segments.

        Args:
        ----
            verbose (bool, optional): to print (or not) results in human
                                      readable format. Defaults to True.
            n_digits (int, optional): to round minute ventilation to n_digits
                                      significant digits. Defaults to 3.
            save (bool, optional): to memorise or not the value obtained.
                                   Defaults to False.

        Raises:
        ------
            ValueError: The frequency and tidal volume must have been
                        set by calling up the corresponding methods.

        Returns:
        -------
            pos_mv (float): minute ventilation of positive segments
                            (in unit of tidal volume .min-1).
            neg_mv (float): minute ventilation of negative segments
                            (in unit of tidal volume .min-1).

        Note:
        ----
            Minute Ventilation is the product of frequency and tidal volume.

        """
        if self.frequency and self.positive_auc:
            pos_mv = scientific_round(
                (self.frequency * self.positive_auc[0]), n_digits=n_digits
            )
            if verbose:
                print(f"Minute ventilation of positive segments: {pos_mv}.")
        else:
            raise ValueError(
                "Please save the frequency and tidal volume by setting the "
                "'save' argument to 'True' in the method call. "
                "(get_frequency() and get_positive_auc())"
            )

        if self.frequency and self.negative_auc:
            neg_mv = scientific_round(
                (self.frequency * self.negative_auc[0]), n_digits=n_digits
            )
            if verbose:
                print(f"Minute ventilation of negative segments: {neg_mv}.")
        else:
            raise ValueError(
                "Please save the frequency and tidal volume by setting the "
                "'save' argument to 'True' in the method call. "
                "(get_frequency() and get_negative_auc())"
            )

        if save:
            self._positive_minute_ventilation = pos_mv
            self._negative_minute_ventilation = neg_mv

        return pos_mv, neg_mv

    @enforce_type_arg(stat=str, kde_min=bool, output_path=str)
    def plot_distribution(self, stat="probability", kde_min=False, output_path=""):
        """
        To get distribution of each feature of the 'BreathingFlow' object.

        Args:
        ----
            stat (int, optional): aggregate statistic to compute in each bin.
                                  Defaults to "probability".
            kde_min (bool, optionnal): to highlight the minimum of the kernel
                                       density estimation. Defaults to False.
            output_path (str, optional): to choose where to save the figure,
                                         if applicable. Defaults to "" (figure not saved)

        Returns:
        -------
            None. Plots the distribution.

        Note:
        ----
            stat should be 'count', 'frequency', 'probability', 'percent' or
            'density'.

        """
        if stat not in ("count", "frequency", "probability", "percent", "density"):
            raise ValueError(
                "stat should be either 'count', 'frequency', 'probability', "
                f"'percent' or 'density'. Not {stat}."
            )

        visualization.plot_features_distribution(
            self.get_positive_time(return_mean=False),
            self.get_negative_time(return_mean=False),
            self.get_positive_auc(return_mean=False),
            self.get_negative_auc(return_mean=False),
            identifier=self.identifier,
            stat=stat,
            kde_min=kde_min,
            output_path=output_path,
        )

    @enforce_type_arg(time_delay=float, color_scheme=str, output_path=str)
    def plot_phase_portrait(self, time_delay=-1.0, color_scheme="time", output_path=""):
        """
        To plot the phase portrait of the air flow rate.

        Args:
        ----
            time_delay (float, optional):
                parameter for phase portrait offset: y(x) vs. y(x+t). Defaults to -1.0.
                The default value -1 is then converted into a tailor-made value.
            color_scheme (str, optional):
                whether the color is defined from time or respiratory phases. Defaults to "time".
            output_path (str, optional): to choose where to save the figure,
                                         if applicable. Defaults to "" (figure not saved)

        Returns:
        -------
            None. Plots the phase portrait.

        """
        if time_delay == -1.0:
            time_delay = (
                self.get_positive_time(verbose=False)[0] +
                self.get_negative_time(verbose=False)[0]) * 1/15

        visualization.plot_phase_portrait(
            x=self.time,
            y=self.flow,
            identifier=self.identifier,
            time_delay=time_delay,
            hz=self.get_hz(),
            color_scheme=color_scheme,
            output_path=output_path
        )

    @enforce_type_arg(as_dict=bool, output_directory=str)
    def get_overview(self, as_dict=False, output_directory=""):
        """
        To summarize the features of the 'BreathingFlow' object in a DataFrame.

        Args:
        ----
            as_dict (bool, optional): whether or not to return the data in
                                      dictionary form. Default to False.
            output_directory (str, optional): where to save the backup file.
                                              It should not be the full path
                                              but just a path to a directory.
                                              Defaults to "" (no backup).

        Returns:
        -------
            pandas.DataFrame: dataframe summarising the features.
            OR
            dict: dictionary summarising the features.

        """
        metrics = ["mean", "std", "n cycle(s)", "variability (%)"]
        dict_data = {}
        dict_data["Bf (rpm)"] = {
            "mean": self._frequency,
            "std": "-",
            "n cycle(s)": "-",
            "variability (%)": "-"
        }
        dict_data["time (AUC > 0) (s)"] = dict(zip(metrics, self.positive_time))
        dict_data["time (AUC < 0) (s)"] = dict(zip(metrics, self.negative_time))
        dict_data["AUC value (AUC > 0)"] = dict(zip(metrics, self.positive_auc))
        dict_data["AUC value (AUC < 0)"] = dict(zip(metrics, self.negative_auc))
        dict_data["Min. ventilation (AUC > 0)"] = {
            "mean": self._positive_minute_ventilation
        }
        dict_data["Min. ventilation (AUC < 0)"] = {
            "mean": self._negative_minute_ventilation
        }

        formatted_dataframe = to_dataframe(
            identifier=self.identifier,
            overview_dict=dict_data
        )

        df_info = self.get_info(shape="df")

        if output_directory:
            backup_dir = os.path.join(output_directory, self.identifier)
            os.makedirs(backup_dir, exist_ok=True)
            if len(self.identifier) > 26:
                identifier = self.identifier[:26]
            excel_path = os.path.join(backup_dir, f"overview_{self.identifier}")

            with pd.ExcelWriter(f"{excel_path}.xlsx", engine="xlsxwriter") as w:
                formatted_dataframe.to_excel(w, sheet_name=f"data_{identifier}")
                df_info.to_excel(w, sheet_name=f"info_{identifier}", index=False)

            ext = f"_{self.identifier}.pdf"
            self.plot(
                show_auc=True,
                output_path=os.path.join(backup_dir, f"flow{ext}")
            )

            self.plot_distribution(
                output_path=os.path.join(backup_dir, f"feat_distrib{ext}")
            )

            self.plot_phase_portrait(
                color_scheme="phases",
                output_path=os.path.join(backup_dir, f"phase_portrait{ext}")
                )

        if as_dict:
            return dict_data
        return formatted_dataframe



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script handling 'BreathingSignals' object: collection of air flow and movements.

Created on Mon Sep  8 14:26:44 2025
@author: CoustilletT
"""


import numpy as np
import os
import pandas as pd

from .breathingflow import BreathingFlow
from .breathingmovement import BreathingMovement
from .coherence import coherence
from .featureextraction import extract_local_minima
from .utils import _check_type, enforce_type_arg, ComparableMixin, to_dataframe
from .visualization import plot_movements


class BreathingSignals(ComparableMixin):
    """Object containing air flow and associated breathing movements."""

    def __init__(self, flow, thorax, abdomen):
        _check_type(flow, BreathingFlow, "flow", allow_none=True)
        _check_type(thorax, BreathingMovement, "thorax")
        _check_type(abdomen, BreathingMovement, "abdomen")

        identifiers = [thorax.identifier, abdomen.identifier]
        if flow is not None:
            identifiers.append(flow.identifier)
        if not all(i == identifiers[0] for i in identifiers):
            raise ValueError(
                f"Cannot combine data with different identifiers: {identifiers}"
            )

        if flow is not None and not np.array_equal(flow.raw_time, thorax.time):
            raise ValueError(
                "The time vectors of the three signals must match together. "
                "'flow.raw_time' does not match 'thorax.time'."
            )

        if not np.array_equal(thorax.time, abdomen.time):
            raise ValueError(
                "The time vectors of the three signals must match together. "
                "'thorax.raw_time' does not match 'abdomen.time'."
            )

        self.flow = flow
        self.thorax = thorax
        self.abdomen = abdomen

        self.identifier = self.thorax.identifier

        self._thorax_freq = self.thorax.get_frequency()

        self._truncate_movements()

    def _truncate_movements(self):
        """To force breathing movements to start and end with a local minimum."""
        first_min, last_min = extract_local_minima(self.thorax.movements)
        self.thorax = self.thorax[first_min:last_min]
        self.abdomen = self.abdomen[first_min:last_min]

    @enforce_type_arg(shape=str)
    def get_info(self, shape="dict"):
        """To get signal information (identifier, start, end and duration.)"""
        return self.thorax.get_info(shape=shape)

    @enforce_type_arg(overlay=bool, output_path=str)
    def plot(self, overlay=False, output_path=""):
        """
        To plot air flow rate and breathing movements on the same plot.

        Args:
        ----
            overlay (bool, optional): whether or not to superimpose respiratory movements.
                                      Defaults to False.

        Returns:
        -------
            None. Plots air flow rate and breathing movements.

        """
        plot_movements(
            y1=self.flow,
            y2=self.thorax,
            y3=self.abdomen,
            overlay=overlay,
            output_path=output_path
        )

    @enforce_type_arg(
        segment_duration=float, output_path=str, view=bool, return_vals=bool
    )
    def get_coherence(
            self, segment_duration=-1.0, output_path="", view=True,
            return_vals=False
    ):
        """To plot the coherence between the breathing movements."""
        # Default value = duration of a respiratory cycle (period)
        if segment_duration == -1.0:
            segment_duration = 60 / self._thorax_freq

        return coherence(
            movement_1=self.thorax,
            movement_2=self.abdomen,
            segment_duration=segment_duration,
            output_path=output_path,
            view=view,
            return_vals=return_vals
        )

    @enforce_type_arg(as_dict=bool, output_directory=str)
    def get_overview(self, as_dict=False, output_directory=""):
        """
        To summarize the features of the 'BreathingSignal' object.

        Args:
        ----
            as_dict (bool, optional): whether or not to return the data in
                                      dictionary form. Default to False.
            output_directory (str, optional): where to save the backup file.
                                              It should not be the full path
                                              but just a path to a directory.
                                              Defaults to "" (no backup).

        Returns:
        -------
            pandas.DataFrame: dataframe summarising the features.
            OR
            dict: dictionary summarising the features.

        """
        overview = {}

        if self.flow is not None:
            overview = self.flow.get_overview(as_dict=True)

        overview["movement coherence (%)"] = self.get_coherence(
            view=False, return_vals=True
        )
        formatted_dataframe = to_dataframe(
            identifier=self.identifier, overview_dict=overview
        )

        df_info = self.thorax.get_info(shape="df")

        if output_directory:
            backup_dir = os.path.join(output_directory, self.identifier)
            os.makedirs(backup_dir, exist_ok=True)
            excel_path = os.path.join(
                backup_dir, f"overview_{self.identifier}"
            )
            if len(self.identifier) > 26:
                identifier = self.identifier[26:]
            with pd.ExcelWriter(f"{excel_path}.xlsx", engine="xlsxwriter") as w:
                formatted_dataframe.to_excel(w, sheet_name=f"data_{identifier}")
                df_info.to_excel(w, sheet_name=f"info_{identifier}", index=False)

            ext = f"_{self.identifier}.pdf"
            self.plot(output_path=os.path.join(backup_dir, f"movements{ext}"))
            self.get_coherence(
                view=False,
                output_path=os.path.join(backup_dir, f"coherence{ext}")
            )

            if self.flow is not None:
                ext = f"_{self.identifier}.pdf"
                self.flow.plot(
                    show_auc=True,
                    output_path=os.path.join(backup_dir, f"flow{ext}")
                )

                self.flow.plot_distribution(
                    output_path=os.path.join(backup_dir, f"feat_distrib{ext}")
                )

                self.flow.plot_phase_portrait(
                    color_scheme="phases",
                    output_path=os.path.join(backup_dir, f"phase_portrait{ext}")
                    )

        if as_dict:
            return overview

        return formatted_dataframe



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script hosting functions to assess the synchronisation of respiratory muscles.

Created on Thu Sep 11 08:38:08 2025
@author: CoustilletT
"""


import numpy as np
from numpy import ma
from scipy.signal import welch, csd

from .visualization import plot_phase_difference, PI


def coherence_score(phase_diff, v1, v2=None):
    """
    To quantify the phase difference between movements.

    Args:
    ----
        phase_diff (list): values of phase shifts.
        v1 (float): lower bound for quantifying phase shifts.
        v2 (float, optional): upper bound for quantifying phase shifts.
                              Defaults to None.

    Returns:
    -------
        float: percentage of values within the range of specified differences.

    """
    N = len(phase_diff)
    if v2 is not None:
        score = round(
            (ma.masked_outside(phase_diff, v1, v2).count()
             + ma.masked_outside(phase_diff, -v1, -v2).count()) / N * 100
        )
        return score
    return round(ma.masked_outside(phase_diff, -v1, v1).count() / N * 100)


def coherence(
        movement_1, movement_2, segment_duration, output_path, view, return_vals
 ):
    """
    To get the coherence of the two movements: whether they are synchronised or not.

    Args:
    ----
        movement_1 (BreathingMovement): Thorax movements.
        movement_2 (BreathingMovement): Abdominal movements.
        segment_duration (float): the duration over which coherence is calculated.
        output_path (str): to choose where to save the figure, if applicable.
        view (bool): whether or not to display the figure.
        return_vals (bool): whether or not to return the % spent in each phase.

    Returns:
    -------
        None. Plots the figure.

    """
    time = movement_1.absolute_time
    y1 = movement_1.movements
    y2 = movement_2.movements
    hz = movement_1.get_hz()

    segment_samples = int(segment_duration * hz)

    segment_indices = []
    phase_diff = []

    for start in range(0, len(time), segment_samples):
        end = start + segment_samples
        if end > len(time):
            break

        mov_1 = y1[start:end]
        mov_2 = y2[start:end]

        # The two movements are assumed to have the same frequency.
        f, Pxx_den = welch(x=mov_1, fs=hz, nperseg=segment_samples)
        dominant_freq = f[np.argmax(Pxx_den)]

        f, Pxy = csd(x=mov_1, y=mov_2, fs=hz, nperseg=segment_samples)
        phase = np.angle(Pxy)

        freq_idx = np.argmin(np.abs(f - dominant_freq))
        segment_phase = phase[freq_idx]

        segment_indices.append((start+end)/(2*hz))
        phase_diff.append(segment_phase)

    # Coherence score.
    normal = coherence_score(phase_diff, v1=PI/6)
    disrupted = coherence_score(phase_diff, v1=PI/2, v2=PI/6)
    asynchronous = coherence_score(phase_diff, v1=5*PI/6, v2=PI/2)
    paradoxical = coherence_score(phase_diff, v1=7*PI/6, v2=5*PI/6)

    coherence_overview = {
        "normal": normal,
        "disrupted": disrupted,
        "asynchronous": asynchronous,
        "paradoxical": paradoxical
    }

    if view:
        print("; ".join(f"{k}: {v}%" for k, v in coherence_overview.items()))

    plot_phase_difference(
        time=time,
        y1=movement_1,
        y2=movement_2,
        segment_indices=segment_indices,
        phase_diff=phase_diff,
        output_path=output_path,
        view=view
    )

    if return_vals:
        return coherence_overview
